@@ -6,7 +6,7 @@ from enum import Enum
 from functools import partial
 from inspect import Parameter, _empty, signature
 from pathlib import Path
-from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union, cast
+from typing import Any, Dict, Generic, List, NamedTuple, Optional, Tuple, Type, TypeVar, cast
 
 try:
     from typing import get_args  # type: ignore
@@ -15,21 +15,29 @@ except ImportError:
 
 import pandas as pd
 from dataclasses_json import dataclass_json
-from flytekit.core.tracker import TrackedInstance
 from flytekit.extras.sqlite3.task import SQLite3Task
 from sklearn.model_selection import train_test_split
 
 import unionml.type_guards as type_guards
 from unionml.defaults import DEFAULT_RESOURCES
+from unionml.tracker import TrackedInstance
 from unionml.utils import inner_task
 
-R = TypeVar("R")  # raw data
-D = TypeVar("D")  # model-ready data
+DT = TypeVar("DT")  # dataset type
+FT = TypeVar("FT")  # feature type
+
+
+class FeatureTypeUnion(Generic[DT, FT]):
+    ...
 
 
 class ReaderReturnTypeSource(Enum):
+    """Indicates which :class:`~unionml.dataset.Dataset` component the materialized dataset type is derived from."""
+
     READER = "reader"
+    """The materialized dataset type comes from the return annotation of :meth:`~unionml.dataset.Dataset.reader`"""
     LOADER = "loader"
+    """The materialized dataset type comes from the return annotation of :meth:`~unionml.dataset.Dataset.loader`"""
 
 
 class Dataset(TrackedInstance):
@@ -394,23 +402,26 @@ class Dataset(TrackedInstance):
         The fallback behavior occurs if the user didn't define a ``feature_transformer`` function.
         """
 
-        parser_type = (
+        dataset_type = (
             # use the reader/loader datatype if parser is the default parser, otherwise use parser return type
             self.dataset_datatype["data"]
             if self._parser == self._default_parser
             else self.parser_return_types[self._parser_feature_key]
         )
 
-        ft_type = (
+        feature_type_ = (
             signature(self._feature_loader).return_annotation
             if self._feature_transformer == self._default_feature_transformer
             else signature(self._feature_transformer).return_annotation
         )
 
-        if parser_type != ft_type:
-            return cast(Type, Union[ft_type, parser_type])
+        if self._feature_loader == self._default_feature_loader:
+            return dataset_type
 
-        return parser_type
+        elif dataset_type != feature_type_:
+            return cast(Type, FeatureTypeUnion[dataset_type, feature_type_])  # type: ignore
+
+        return dataset_type
 
     @classmethod
     def _from_flytekit_task(
@@ -458,7 +469,7 @@ class Dataset(TrackedInstance):
         """
         return cls._from_flytekit_task(task, *args, **kwargs)
 
-    def _default_loader(self, data: R) -> R:
+    def _default_loader(self, data: Any) -> Any:
         [(_, data_type)] = self.dataset_datatype.items()
         if data_type is pd.DataFrame:
             return pd.DataFrame(data)
@@ -466,21 +477,21 @@ class Dataset(TrackedInstance):
 
     def _default_splitter(
         self,
-        data: D,
+        data: Any,
         test_size: float,
         shuffle: bool,
         random_state: int,
-    ) -> Tuple[D, ...]:
+    ) -> Tuple[Any, ...]:
         if not isinstance(data, pd.DataFrame):
             return (data,)
         return train_test_split(data, test_size=test_size, random_state=random_state, shuffle=shuffle)
 
     def _default_parser(
         self,
-        data: D,
+        data: Any,
         features: Optional[List[str]],
         targets: Optional[List[str]],
-    ) -> Tuple[D, D]:
+    ) -> Tuple[Any, Any]:
         if not isinstance(data, pd.DataFrame):
             return (data,)  # type: ignore
 
@@ -492,7 +503,7 @@ class Dataset(TrackedInstance):
             target_data = pd.DataFrame()
         return data[features], target_data
 
-    def _default_feature_loader(self, features: Any) -> R:
+    def _default_feature_loader(self, features: Any) -> Any:
         if isinstance(features, Path):
             with features.open() as f:
                 features = json.load(f)
@@ -508,9 +519,9 @@ class Dataset(TrackedInstance):
 
         return features
 
-    def _default_feature_transformer(self, features: R) -> D:
+    def _default_feature_transformer(self, features: Any) -> Any:
         """
         By default this is just a pass-through function. The user can choose to override it with
         the `@dataset.feature_transformer` decorator.
         """
-        return cast(D, features)
+        return features
